syntax = "proto3";

package nf.common;

// When representing robot world space, Z is up
message Vec3 {
  float x = 1;
  float y = 2;
  float z = 3;
}

message Pose {
  Vec3 rotation = 1; // Rodriguez rotation vector
  Vec3 position = 2;
}

// Episode control commands

enum EpCommand {
  EPCOMMAND_NOTHING = 0;
  EPCOMMAND_ABANDON = 1;
  EPCOMMAND_END_RECORDING = 2;
  EPCOMMAND_START_OR_COMPLETE = 3;
}

message EpisodeControl {
  EpCommand command = 1;
}